
import itsdangerous
from restfulpy.principal import JwtPrincipal
from restfulpy.orm import DBSession

from nanohttp import Controller, context, settings, json, RestController


class JwtController(Controller):
    token_key = 'HTTP_AUTHORIZATION'

    def begin_request(self):
        if self.token_key in context.environ:
            try:
                context.identity = JwtPrincipal.decode(context.environ[self.token_key])
            except itsdangerous.BadData:
                context.identity = None
        else:
            context.identity = None

    # noinspection PyMethodMayBeStatic
    def begin_response(self):
        if settings.debug:
            context.response_headers.add_header('Access-Control-Allow-Origin', '*')
            context.response_headers.add_header('Access-Control-Allow-Methods',
                                                'GET, POST, PUT, DELETE, UNDELETE, METADATA, PATCH')
            context.response_headers.add_header('Access-Control-Allow-Headers', 'Content-Type, Authorization')
            context.response_headers.add_header('Access-Control-Expose-Headers',
                                                'Content-Type, X-Pagination-Count, X-Pagination-Skip, '
                                                'X-Pagination-Take')

    # noinspection PyMethodMayBeStatic
    def end_response(self):
        DBSession.remove()

    def __call__(self, *remaining_paths):
        if context.method == 'options':
            context.response_headers.add_header("Cache-Control", "no-cache,no-store")
            return b''

        return super(JwtController, self).__call__(*remaining_paths)


class ModelRestController(RestController):
    __model__ = None

    @json
    def metadata(self):
        return self.__model__.json_metadata()





class RestfulException(Exception):
    def to_json(self):
        return {
            'type': str(type(self)),
            'content': str(self),
        }


class FileNotFoundException(RestfulException):
    pass


class OrmException(RestfulException):
    pass


class KeyDecodeError(OrmException):
    pass


class FormParserError(RestfulException):
    def __init__(self):
        super(FormParserError, self).__init__('Cannot parse the form.')


class ValidationError(RestfulException):
    field = None

    def __init__(self, field_name, *args, **kw):
        self.field = field_name
        super(ValidationError, self).__init__(*args, **kw)

    def to_json(self):
        d = super(ValidationError, self).to_json()
        d['field'] = self.field
        return d




from urllib.parse import urlparse

from sqlalchemy import create_engine
from nanohttp import settings


class DatabaseManager(object):

    def __init__(self):

        self.db_uri = settings.db.uri
        self.db_name = urlparse(self.db_uri).path.lstrip('/')
        self.admin_uri = settings.db.administrative_uri
        self.admin_db_name = urlparse(settings.db.administrative_uri).path.lstrip('/')
        if self.db_uri.startswith('sqlite'):
            raise NotImplementedError("Sqlite database management is not supported yet.")

    def __enter__(self):
        self.engine = create_engine(self.admin_uri)
        self.connection = self.engine.connect()
        self.connection.execute('commit')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connection.close()
        self.engine.dispose()

    def exists_database(self):
        r = self.connection.execute('SELECT 1 FROM pg_database WHERE datname = \'%s\'' % self.db_name)
        try:
            ret = r.cursor.fetchall()
            return ret
        finally:
            r.cursor.close()

    def create_database(self):
        self.connection.execute('CREATE DATABASE %s' % self.db_name)
        self.connection.execute('commit')

    def create_database_if_not_exists(self):
        if not self.exists_database():
            self.create_database()

    def drop_database(self):
        self.connection.execute('DROP DATABASE IF EXISTS %s' % self.db_name)
        self.connection.execute('commit')



from os import makedirs
from os.path import join, exists, dirname, basename
from urllib.parse import quote
import warnings

from webtest import TestApp

from restfulpy.testing.constants import DOC_HEADER, DOC_LEGEND


class RequestSignature(object):
    def __init__(self, role, method, url, query_string=None, request_headers=None, response_headers=None):
        self.role = role
        self.method = method
        self.url = url
        self.query_string = query_string
        self.request_headers = request_headers
        self.response_headers = response_headers

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        return hash((
            self.role, self.method, self.url, self.query_string,
            tuple(self.request_headers.items()) if self.request_headers else None,
            tuple(self.response_headers.items()) if self.response_headers else None
        ))


class FormParameter(object):
    def __init__(self, name, value=None, type_=str, optional=False, default=None):
        self.name = name
        self.type_ = type_
        self.optional = optional
        self.value = value
        self.default = default

    @property
    def type_string(self):
        if self.type_ is None:
            return ''
        return self.type_ if isinstance(self.type_, str) else self.type_.__name__

    @property
    def value_string(self):
        if self.value is None:
            return ''

        if self.type_ == 'file':
            return basename(self.value)

        if self.type_ is bool:
            return str(self.value).lower()

        return self.value

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        return hash(self.name)


class DocumentaryTestApp(TestApp):
    _files = []
    _signatures = set()
    _jwt_token = None
    __jwt_header_key__ = 'HTTP_AUTHORIZATION'

    def __init__(self, destination_dir, application, *args, **kwargs):
        self.application = application
        self.destination_dir = destination_dir
        super(DocumentaryTestApp, self).__init__(application.wsgi(), *args, **kwargs)

    @property
    def legend_filename(self):
        return join(self.destination_dir, 'legend.md')

    @property
    def jwt_token(self):
        return self._jwt_token

    @jwt_token.setter
    def jwt_token(self, value):
        self._jwt_token = value

        if value is not None:
            self.extra_environ.update({
                self.__jwt_header_key__: value,
            })
        else:
            if self.__jwt_header_key__ in self.extra_environ:
                del self.extra_environ[self.__jwt_header_key__]

    def _ensure_file(self, filename, entity):
        d = dirname(filename)
        if not exists(d):
            makedirs(d)

        if not exists(self.legend_filename):
            with open(self.legend_filename, 'w') as f:
                f.write(DOC_LEGEND)

        if filename in self._files:
            return open(filename, 'a')
        else:
            f = open(filename, 'w')
            f.write(DOC_HEADER % dict(version=self.application.version))
            f.write('\n%s' % entity)
            f.write('\n%s\n' % ('-' * len(entity)))
            self._files.append(filename)
            return f

    def document(self, role, method, url, resp, request_headers, model=None, params=None, query_string=None):
        signature = RequestSignature(
            role, method, url,
            query_string=tuple(query_string.keys()) if query_string else None,
            request_headers=request_headers,
            response_headers=resp.headers
        )
        if signature in self._signatures:
            return
        path_parts = url.split('?')[0].split('/')[1:]
        if len(path_parts) == 1:
            p = path_parts[0].strip()
            entity = filename = p if p else 'index'
        else:
            version, entity = path_parts[0:2]
            filename = '_'.join(path_parts[:2] + [method.lower()])

        filename = join(self.destination_dir, '%s.md' % filename)
        f = self._ensure_file(filename, entity)

        # Extracting more params & info from model if available
        if params and model:
            for c in model.iter_json_columns(relationships=False, include_readonly_columns=False):
                json_name = c.info['json']
                column = model.get_column(c)

                if hasattr(column, 'default') and column.default:
                    default_ = column.default.arg if column.default.is_scalar else 'function(...)'
                else:
                    default_ = ''

                if 'attachment' in column.info:
                    type_ = 'attachment'
                else:
                    type_ = str if 'unreadable' in column.info and column.info['unreadable'] else \
                        column.type.python_type

                if json_name in params:
                    param = params[params.index(json_name)]
                    param.default = default_
                    if param.type_ is None:
                        param.type_ = type_
                    if param.optional is None:
                        param.optional = column.nullable
                else:
                    param = FormParameter(
                        json_name,
                        type_=type_,
                        optional=column.nullable,
                        default=default_)
                    params.append(param)

        try:
            f.write('\n- (%s) **%s** `%s`\n' % (role, method.upper(), url))
            if params:
                f.write('\n    - Form Parameters:\n\n')
                f.write('        | Parameter | Optional | Type | Default | Example |\n')
                f.write('        | --------- | -------- | ---- | ------- | ------- |\n')
                for param in params:
                    f.write('        | %s | %s | %s | %s | %s |\n' % (
                        param.name,
                        True if method.lower() == 'put' else param.optional,
                        param.type_string,
                        param.default if param.default is not None else '',
                        param.value_string))
            if query_string:
                f.write('\n    - Query String:\n\n')
                f.write('        | Parameter | Example |\n')
                f.write('        | --------- | ------- |\n')
                for name, value in query_string.items():
                    f.write('        | %s | %s |\n' % (
                        name,
                        str(value)))

            if request_headers:
                f.write('\n    - Request Headers:\n\n')
                for k, v in request_headers.items():
                    f.write('%s%s: %s\n' % (12 * ' ', k, v))

            f.write('\n    - Response Headers:\n\n')
            for k, v in resp.headers.items():
                f.write('%s%s: %s\n' % (12 * ' ', k, v))

            f.write('\n    - Response Body:\n\n')
            for l in resp.body.decode().splitlines():
                f.write('%s%s\n' % (12 * ' ', l))
            f.write('\n\n')
            self._signatures.add(signature)
        finally:
            f.write('\n')
            f.close()

    def send_request(self, role, method, url, query_string=None, url_params=None,
                     params=None, model=None, doc=True, **kwargs):
        files = []
        parameters = {}
        if params:
            if isinstance(params, dict):
                parameters = params
                if doc:
                    warnings.warn('Skipping documentation generation, because the passed parameters are plain dict.')
                    doc = False
            else:
                for param in params:
                    if param.type_ == 'file':
                        files.append((param.name, param.value))
                    else:
                        parameters[param.name] = param.value

        if query_string:
            parameters.update(query_string)

        if files:
            kwargs['upload_files'] = files
        if parameters:
            kwargs['params'] = parameters

        real_url = (url % url_params) if url_params else url
        real_url = quote(real_url)
        kwargs['expect_errors'] = True
        resp = getattr(self, method.lower())(real_url, **kwargs)

        if doc:
            self.document(role, method, url, resp, kwargs.get('headers'), model=model, params=params, query_string=query_string)
        return resp

    def metadata(self, url, params='', headers=None, extra_environ=None,
                 status=None, upload_files=None, expect_errors=False,
                 content_type=None, xhr=False):
        if xhr:
            headers = self._add_xhr_header(headers)
        return self._gen_request('METADATA', url, params=params, headers=headers,
                                 extra_environ=extra_environ, status=status,
                                 upload_files=upload_files,
                                 expect_errors=expect_errors,
                                 content_type=content_type,
                                 )

    def undelete(self, url, params='', headers=None, extra_environ=None,
                 status=None, upload_files=None, expect_errors=False,
                 content_type=None, xhr=False):
        if xhr:
            headers = self._add_xhr_header(headers)
        return self._gen_request('UNDELETE', url, params=params, headers=headers,
                                 extra_environ=extra_environ, status=status,
                                 upload_files=upload_files,
                                 expect_errors=expect_errors,
                                 content_type=content_type,
                                 )



import argparse
import sys

import argcomplete

from restfulpy.cli.admin import AdminLauncher
from restfulpy.cli.base import Launcher
from restfulpy.cli.migrate import MigrateLauncher
from restfulpy.cli.serve import ServeLauncher
from restfulpy.cli.worker import WorkerLauncher


class MainLauncher(Launcher):

    def __init__(self, application):
        self.application = application
        self.parser = parser = argparse.ArgumentParser(description='%s command line interface.' % application.name)
        parser.add_argument('-c', '--config-file', metavar="FILE",
                            help='List of configuration files separated by space. Default: ""')
        parser.add_argument('-d', '--config-dir', metavar="DIR",
                            help='List of configuration directories separated by space. Default: ""')
        subparsers = parser.add_subparsers(title="sub commands", prog=application.name, dest="command")

        AdminLauncher.register(subparsers)
        ServeLauncher.register(subparsers)
        MigrateLauncher.register(subparsers)
        WorkerLauncher.register(subparsers)
        application.register_cli_launchers(subparsers)
        argcomplete.autocomplete(parser)

    def launch(self, args=None):
        if args:
            cli_args = self.parser.parse_args(args)
        else:
            cli_args = self.parser.parse_args()

        cli_args.application = self.application
        self.application.configure(files=cli_args.config_file)
        self.application.initialize_models()
        cli_args.func(cli_args)
        sys.exit(0)

    @classmethod
    def create_parser(cls, subparsers):
        """
        Do nothing here
        """
        pass



import base64

from itsdangerous import TimedJSONWebSignatureSerializer
from nanohttp import settings


class JwtPrincipal:

    def __init__(self, payload):
        self.payload = payload

    @classmethod
    def create_serializer(cls):
        return TimedJSONWebSignatureSerializer(
            settings.jwt.secret,
            expires_in=settings.jwt.max_age,
            algorithm_name=settings.jwt.algorithm
        )

    def encode(self):
        return self.create_serializer().dumps(self.payload)

    @classmethod
    def decode(cls, encoded):
        if encoded.startswith('Bearer '):
            encoded = encoded[7:]
        if encoded.startswith('Basic '):
            encoded = base64.decodebytes(encoded[6:].encode())
        payload = cls.create_serializer().loads(encoded)
        return cls(payload)

    def is_in_roles(self, *roles):
        if 'roles' in self.payload:
            if set(self.payload['roles']).intersection(roles):
                return True
        return False

    @property
    def email(self):
        return self.payload.get('email')

    @property
    def session_id(self):
        return self.payload.get('sessionId')

    @property
    def id(self):
        return self.payload.get('id')
